// Authors: Korbinian Schneeberger, Joerg Hagmann and Sandra Gesing
// Copyright (C) 2008 by Max-Planck Institute for Developmental Biology, Tuebingen, Germany

#include "genomemapper.h"

#include <limits.h>


int init_defaults();
int init_options(int argc, char *argv[]);
int init_chr_index_file();
int init_meta_index_file();
int init_constants();
int init_query_file();
int init_output_file();
int init_spliced_output_file();
int init_leftover_file();
int init_genome_file();
int init_hit_lists();
int init_operators();
//int init_statistic_vars();	
int init_alignment_structures();

int init(int argc, char *argv[]) 
{
	init_defaults();
	init_options(argc, argv);
	init_chr_index_file();
	init_meta_index_file();
	init_query_file();
	init_output_file();
	init_spliced_output_file();
	if (strlen(LEFTOVER_FILE_NAME) > 0) init_leftover_file();
	init_genome_file();
	init_alignment_structures();
	init_hit_lists();

	if (MAX_OCC_PER_SEED != DEFAULT_MAX_OCC_PER_SEED || SEED_EXTRA_EXTEND_THRESHOLD != DEFAULT_MAX_OCC_PER_SEED)
		NOT_MAXIMAL_HITS = 1;

	if (EXTRA_SEED_EXTEND == 0 && SEED_EXTRA_EXTEND_THRESHOLD != DEFAULT_MAX_OCC_PER_SEED)
		SEED_EXTRA_EXTEND_THRESHOLD = DEFAULT_MAX_OCC_PER_SEED;

	return 0;
}

int init_alignment_structures() 
{
#ifdef THREADS 
        MAX_READ_LENGTH=2000;
#else
        MAX_READ_LENGTH=21;
#endif
	/////////////////////////////
	////// if you change NUM_EDIT_OPS here, then put alloc_hits_by_editops in init_hit_lists_operator() after this function!!!
	/////////////////////////////
	
	if (NUM_MISMATCHES > NUM_EDIT_OPS) {
		fprintf(stderr, "\n!!! WARNING: Nr. of allowed mismatches is set to nr. of allowed edit operations!\n\n");
		NUM_MISMATCHES = NUM_EDIT_OPS;
	}
	if (NUM_GAPS > NUM_EDIT_OPS) {
		fprintf(stderr, "\n!!! WARNING: Nr. of allowed gaps is set to nr. of allowed edit operations!\n\n");
		NUM_GAPS = NUM_EDIT_OPS;
	}

	if (M_SCORE == MM_SCORE || M_SCORE == GAP_SCORE) {
		fprintf(stderr, "ERROR: Sorry, until now, program cannot handle equal mismatch-/gap- and match-scores! Please restart with different values!\n");
		exit(1);
	}
	
	if (VERBOSE) printf("allowed edit operations: %d / allowed mismatches: %d / allowed gaps: %d\n%s strategy\t%s overhang alignment\n------------------\n", 
		NUM_EDIT_OPS, NUM_MISMATCHES, NUM_GAPS,	ALL_HIT_STRATEGY? "all hit": SECOND_BEST_HIT_STRATEGY? "second best hit": "best hit", OVERHANG_ALIGNMENT? "with": "without");
	
	
	if (GAP_SCORE <= MM_SCORE) {
		fprintf(stderr, "\nWARNING: It is not recommended to set gap score smaller than mismatch score. This reduces runtime significantly.\n\n");
		WORST_SCORE = NUM_GAPS * GAP_SCORE + (NUM_EDIT_OPS - NUM_GAPS) * MM_SCORE;
	}
	else {
		WORST_SCORE = NUM_GAPS * GAP_SCORE + (NUM_EDIT_OPS - NUM_GAPS) * MM_SCORE;
	}
	WORST_MM_SCORE = NUM_MISMATCHES * MM_SCORE;
	
	// add one byte to allow adding terminating zero
	ALIGNSEQ = (char *) malloc ((MAX_READ_LENGTH + 1 + 4 * MAX_EDIT_OPS ) * sizeof(char));

	if(ALIGNSEQ==NULL) {
		perror("ERROR: could not allocate memory for alignment array");
		exit(1);
	}

#ifdef THREADS 
	
	M = (unsigned int***) malloc (NUM_THREADS * sizeof(unsigned int**));
	T = (char***) malloc (NUM_THREADS * sizeof(char**));

	if((M==NULL)||(T==NULL)) {
		perror("ERROR: could not allocate memory for M ***, T ***");
		exit(1);
	}

	int j, i;
	for (j=0; j!=NUM_THREADS; ++j) {
		M[j] = (unsigned int**) malloc ((NUM_GAPS * 2 + 1) * sizeof(unsigned int*));
		T[j] = (char**) malloc ((NUM_GAPS * 2 + 1) * sizeof(char*));

		if((M[j]==NULL)||(T[j]==NULL)) {
			perror("ERROR: could not allocate memory for M **, T **");
			exit(1);
		}
	
		for (i=0; i!=2*NUM_GAPS+1; ++i) {
			M[j][i] = (unsigned int*) malloc (MAX_READ_LENGTH * sizeof(unsigned int));
			T[j][i] = (char*) malloc (MAX_READ_LENGTH * sizeof(char));

			if((M[j][i]==NULL)||(T[j][i]==NULL)) {
				perror("ERROR: could not allocate memory for M *, T *");
				exit(1);
			}
		}
	}

        thread_data = (THREAD_DATA**) malloc (NUM_THREADS * sizeof(THREAD_DATA*));

	if(thread_data == NULL) {
		perror("ERROR: could not allocate memory for thread data");
		exit(1);
	}

        for (i=0; i < NUM_THREADS; ++i) { 
                thread_data[i] = (THREAD_DATA*) malloc (sizeof(THREAD_DATA));

		if(thread_data[i] == NULL) {
			perror("ERROR: could not allocate memory for thread data entry");
			exit(1);
		}

		thread_data[i]->hitdepth_start = (int*) malloc (MAX_READ_LENGTH * sizeof(int)); 
		thread_data[i]->hitdepth_end = (int*) malloc (MAX_READ_LENGTH * sizeof(int));
		thread_data[i]->thread_const = (THREAD_CONST*) malloc (sizeof(THREAD_CONST));

		if((thread_data[i]->hitdepth_start == NULL)||(thread_data[i]->hitdepth_end)||(thread_data[i]->thread_const)) {
			perror("ERROR: could not allocate memory for thread data entry (2)");
			exit(1);
		}
	}

	THREAD_STARTED = (int*) malloc (NUM_THREADS * sizeof(int));
 	start_align_mutex = (pthread_mutex_t*) malloc (NUM_THREADS * sizeof(pthread_mutex_t));
 	start_align_cond = (pthread_cond_t*) malloc (NUM_THREADS * sizeof(pthread_cond_t));
	
#else 
	M = (unsigned int**) malloc ((2 * NUM_GAPS + 1) * sizeof(unsigned int*));
	T = (char**) malloc ((2 * NUM_GAPS + 1) * sizeof(char*));

	if((M == NULL) || (T == NULL)) {
		perror("ERROR: could not allocate memory for M **, T **");
		exit(1);
	}
	
	int i;
	for (i=0; i!=2*NUM_GAPS+1; ++i) {
		*(M+i) = (unsigned int*) malloc ((MAX_READ_LENGTH + NUM_GAPS * 2) * sizeof(unsigned int));
		*(T+i) = (char*) malloc ((MAX_READ_LENGTH + NUM_GAPS * 2) * sizeof(char));

		if((M[i] == NULL) || (T[i] == NULL)) {
			perror("ERROR: could not allocate memory for M *, T *");
			exit(1);
		}
	}
#endif

	//only for debugging purposes:
	//chrseq = (char *) malloc (MAX_READ_LENGTH * sizeof(char));
	//chrseq[0] = '\0';
	
	return(0);
}

/// Re-initialize alignment structures for a new MAX_READ_LENGTH.
int realloc_alignment_structures() 
{
	//fprintf(stderr,"DEBUG   realloc_alignment_structures %d %d %ld\n",MAX_READ_LENGTH,MAX_EDIT_OPS,(MAX_READ_LENGTH + 3 * MAX_EDIT_OPS) * sizeof(char));
	// add one byte to allow adding terminating zero
	ALIGNSEQ = (char *) realloc (ALIGNSEQ,(MAX_READ_LENGTH + 1 + 4 * MAX_EDIT_OPS) * sizeof(char));

        if(ALIGNSEQ==NULL) return -1;

#ifdef THREADS 
	return -1;
#else 
	int i;
	for (i=0; i!=2*NUM_GAPS+1; ++i) {
		*(M+i) = (unsigned int*) realloc (*(M+i),(MAX_READ_LENGTH + NUM_GAPS * 2) * sizeof(unsigned int));
		*(T+i) = (char*) realloc (*(T+i),(MAX_READ_LENGTH + NUM_GAPS * 2) * sizeof(char));

		if(*(M+i)==NULL) return -1;
		if(*(T+i)==NULL) return -1;
	}
#endif

	return(0);
}

int init_from_meta_index() 
{
	init_constants(); // updated

	alloc_genome_memory(); // updated

	init_operators(); //updated 

	if (!HITLEN_LIMIT) HITLEN_LIMIT = INDEX_DEPTH;

	return(0);
}

int init_defaults() 
{

//SG	
#ifdef THREADS
	NUM_HITS = 0;
	NUM_THREADS = 4;
#endif
#ifdef CUDA
        int DEVICE_COUNT; 
#endif
	//NUM_ALIGNMENTS = 0;

	ALL_HIT_STRATEGY = 0;		// default: best hit strategy
	SECOND_BEST_HIT_STRATEGY = 0;
	HITLEN_LIMIT = 0;		// only temporarily 0
	
	int i;

	VERBOSE = 0;

	MAP_REVERSE = 1;

	POWER[0] = 1;
	for (i = 1 ; i <= MAX_INDEX_DEPTH; i++) {
		POWER[i] = POWER[i-1] * 4;
	}
	
	USR_READ_LENGTH = INT_MAX;
	HAS_SLOT = 0;
	READS_MAPPED = 0;
	NUM_EDIT_OPS = MAX_EDIT_OPS = 4;
	NUM_GAPS = 1;
	NUM_MISMATCHES = 3;
	M_SCORE = 0;
	MM_SCORE = 4;
	GAP_SCORE = 5;
	GAPS_MOST_RIGHT = 0;
	OVERHANG_ALIGNMENT = 1;
	STRINGENT_GAPLIMIT = 1;
	PRINT_SEQ = 0;
	FLANKING = 0;

	READNR = 0;
	FROM_READ = 1;
	
	SPLICED_HITS = 0;
	SPLICED_HIT_MIN_LENGTH_LONG = 15;
	SPLICED_HIT_MIN_LENGTH_SHORT = 8;
	SPLICED_HIT_MIN_LENGTH_COMB = 28;
	SPLICED_HIT_MAX_DIST = 100000;
	SPLICED_LONGEST_HIT_MIN_LENGTH = 25;

	OUTPUT_FORMAT = 0;
	
	SCORES_OUT = 1;
	
	REPEATMAP = 0;
	
	CHROM_CONTAINER_SIZE = 15000000;

	MAX_OCC_PER_SEED = DEFAULT_MAX_OCC_PER_SEED;
	SEED_EXTRA_EXTEND_THRESHOLD = DEFAULT_MAX_OCC_PER_SEED;
	EXTRA_SEED_EXTEND = 0;
	NOT_MAXIMAL_HITS = 0;
	
	//STATISTICS = 0;
	
	//READ_ID = (char *) malloc(MAX_READ_ID_LENGTH * sizeof(char));
	//READ = (char *) malloc(MAX_READ_LENGTH * sizeof(char));
	READ = astr_create(NULL);
        READ_ID = astr_create(NULL);
        READ_PE_FLAG = astr_create(NULL);

        FLANK_SEQ = astr_create(NULL);

	for (i = 0; i != 3; ++i) READ_QUALITY[i] = astr_create(NULL);
	linenr = 0;
	READ_LENGTH = 0;
	
	//listcount = 0;
	//listocc = 0;
	
        return(0);
}

int init_options(int argc, char *argv[]) 
{
	int i;
	char not_defined;
	char has_index = 0;
	char has_query = 0;
	char has_genome = 0;
	char output[5];

	for (i = 1; i < argc; i++) {
		not_defined = 1;

		//genome file
		if(strcmp(argv[i],"-i")==0){

			not_defined = 0;
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(GENOME_FILE_NAME, argv[i]);
			has_genome = 1;
			has_index = 1;

#ifndef METHYLOME
			strcpy(CHR_INDEX_FILE_NAME, argv[i]);
			strcpy(CHR_INDEX_FILE_NAME + strlen(argv[i]), ".cid");
			strcpy(META_INDEX_FILE_NAME, argv[i]);
			strcpy(META_INDEX_FILE_NAME + strlen(argv[i]), ".mta");
			strcpy(INDEX_FWD_FILE_NAME, argv[i]);
			strcpy(INDEX_FWD_FILE_NAME + strlen(argv[i]), ".mfd");
			strcpy(INDEX_REV_FILE_NAME, argv[i]);
			strcpy(INDEX_REV_FILE_NAME + strlen(argv[i]), ".mrc");
#else
			strcpy(CHR_INDEX_FILE_NAME, argv[i]);
                        strcpy(CHR_INDEX_FILE_NAME + strlen(argv[i]), ".mid");
			strcpy(META_INDEX_CT_FILE_NAME, argv[i]);
                        strcpy(META_INDEX_CT_FILE_NAME + strlen(argv[i]), ".mtc");
                        strcpy(INDEX_FWD_CT_FILE_NAME, argv[i]);
                        strcpy(INDEX_FWD_CT_FILE_NAME + strlen(argv[i]), ".cfd");
                        strcpy(INDEX_REV_CT_FILE_NAME, argv[i]);
                        strcpy(INDEX_REV_CT_FILE_NAME + strlen(argv[i]), ".crc");
			strcpy(META_INDEX_GA_FILE_NAME, argv[i]);
                        strcpy(META_INDEX_GA_FILE_NAME + strlen(argv[i]), ".mtg");
                        strcpy(INDEX_FWD_GA_FILE_NAME, argv[i]);
                        strcpy(INDEX_FWD_GA_FILE_NAME + strlen(argv[i]), ".gfd");
                        strcpy(INDEX_REV_GA_FILE_NAME, argv[i]);
                        strcpy(INDEX_REV_GA_FILE_NAME + strlen(argv[i]), ".grc");		
#endif
		}

		//query file
		if(strcmp(argv[i],"-q")==0){
			not_defined = 0;
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(QUERY_FILE_NAME, argv[i]);
			has_query = 1;
		}

    		//output file
		if(strcmp(argv[i],"-o")==0){
			not_defined = 0;
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(OUT_FILE_NAME, argv[i]);
		}
		
		//spliced output file
                if(strcmp(argv[i],"-H")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        strcpy(SPLICED_OUT_FILE_NAME, argv[i]);
                }

		//spliced hit distance
                if(strcmp(argv[i],"-D")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        SPLICED_HIT_MAX_DIST = atoi(argv[i]);
                }

		//spliced hits min combined length
                if(strcmp(argv[i],"-C")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        SPLICED_HIT_MIN_LENGTH_COMB = atoi(argv[i]);
                }

		//spliced hits min length
                if(strcmp(argv[i],"-K")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        SPLICED_HIT_MIN_LENGTH_SHORT = atoi(argv[i]);
                }

		//spliced hits min length of longer hit
                if(strcmp(argv[i],"-L")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        SPLICED_HIT_MIN_LENGTH_LONG = atoi(argv[i]);
                }

		//spliced longest hit min length
		if(strcmp(argv[i],"-N")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        SPLICED_LONGEST_HIT_MIN_LENGTH = atoi(argv[i]);
                }

		//read length limit
		if(strcmp(argv[i],"-k")==0){
                        not_defined = 0;
                        if(i+1 > argc - 1){ usage(); exit(1); }
                        i++;
                        USR_READ_LENGTH = atoi(argv[i]);
			//if (USR_READ_LENGTH > MAX_READ_LENGTH) {
			//	fprintf(stderr, "WARNING: User defined read length is larger than maximal allowed length. It will be disregarded!\n");
			//}
                }

		//output format
		if(strcmp(argv[i],"-f")==0){
			not_defined = 0;
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			printf("argv .%s.\n",argv[i]);
			strcpy(output, argv[i]);
			if (strcmp(output, "shore") == 0 || strcmp(output, "SHORE") == 0) {
				OUTPUT_FORMAT = 0;
			}
			else if (strcmp(output, "bed") == 0 || strcmp(output, "BED") == 0) {
				OUTPUT_FORMAT = 1;
			}
			else {
				fprintf(stderr, "ERROR: Output file format must either be \"shore\" or \"bed\"\n");
				exit(2);
			}
		}
		
		//leftover file
		if(strcmp(argv[i],"-u")==0){
			not_defined = 0;
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(LEFTOVER_FILE_NAME, argv[i]);
		}

		//verbose
		if(strcmp(argv[i],"-v")==0){
			not_defined = 0;
			VERBOSE = 1;
		}
		
		//scores out
		if(strcmp(argv[i],"-e")==0){
			not_defined = 0;
			SCORES_OUT = 0;
		}

		//build reverse index
		if(strcmp(argv[i],"-r")==0){
			not_defined = 0;
			MAP_REVERSE = 0;
		}

		//spliced hits analysis
		if (strcmp(argv[i],"-S") == 0) {
			not_defined = 0;
			SPLICED_HITS = 1;
		}
		
		//report all hits-strategy
		if(strcmp(argv[i],"-a")==0){
			not_defined = 0;
			ALL_HIT_STRATEGY = 1;
		}

		//report second best hits-strategy
		if(strcmp(argv[i],"-b")==0){
			not_defined = 0;
			SECOND_BEST_HIT_STRATEGY = 1;
		}
		
		//max nr of alignments per read
		if(strcmp(argv[i],"-ar")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if (REPEATMAP != 0) {
				fprintf(stderr, "ERROR: options -a and -ar exclude themselves!\n");
				exit(2);
			}
			if ((REPEATMAP = atoi(argv[i])) == 0) {
				if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of alignments must be an integer value!\n");
					exit(2);
				}
			}
		}
		
		//max nr of alignments per read, randomly chosen!
		if(strcmp(argv[i],"-n")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if (REPEATMAP != 0) {
				fprintf(stderr, "ERROR: options -a and -ar exclude themselves!\n");
				exit(2);
			}
			if ((REPEATMAP = atoi(argv[i])) == 0) {
				if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of alignments must be an integer value!\n");
					exit(2);
				}
			}
			REPEATMAP = -REPEATMAP;
		}
		
		//max number of allowed edit operations
		if(strcmp(argv[i],"-E")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((NUM_EDIT_OPS = atoi(argv[i])) == 0) {
				if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of edit operations must be an integer value!\n");
					exit(2);
				}
			}
			//if (NUM_EDIT_OPS > MAX_EDIT_OPS) {
			//	fprintf(stderr, "ERROR: Number of allowed mismatches exceeds maximal number of edit operations (=%d)! Please restart with a lower value!\n", MAX_EDIT_OPS);
			//	exit(2);
			//}
                        MAX_EDIT_OPS = NUM_EDIT_OPS;
		}
		
		//max number of allowed mismatches
		if(strcmp(argv[i],"-M")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((NUM_MISMATCHES = atoi(argv[i])) == 0) {
				 if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of mismatches must be an integer value!\n");
					exit(2);
				}
			}
		}
		
		//max number of allowed gaps
		if(strcmp(argv[i],"-G")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((NUM_GAPS = atoi(argv[i])) == 0) {
				if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of gaps must be an integer value!\n");
					exit(2);
				}
			}
		}
	
		//match score
		if(strcmp(argv[i],"-match_score")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			M_SCORE = atoi(argv[i]);
		}
		
		//mismatch score
		if(strcmp(argv[i],"-m")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			MM_SCORE = atoi(argv[i]);
			//SG
			//if (MM_SCORE < 0) {
			//	fprintf(stderr, "ERROR: Mismatch score must be positive!\n");
			//	exit(2);
			//}
			if (MM_SCORE == 0) fprintf(stderr, "\n!!! WARNING: Mismatch score is 0! This could lead to bad alignments!\n\n");
		}
		
		//gap score
		if(strcmp(argv[i],"-g")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			GAP_SCORE = atoi(argv[i]);
			//SG
			//if (GAP_SCORE < 0) {
			//	fprintf(stderr, "ERROR: Gap score must be positive!\n");
			//	exit(2);
			//}
			if (GAP_SCORE == 0) fprintf(stderr, "\n!!! WARNING: Gap score is 0! This could lead to bad alignments!\n\n");
		}
		
		//hitlength limit
		if(strcmp(argv[i],"-l")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((HITLEN_LIMIT = atoi(argv[i])) == 0) {
				fprintf(stderr, "ERROR: Hitlength limit must be an integer value unequal to 0!\n");
				exit(2);
			}
		}
		
		//chromosome container size
		if(strcmp(argv[i],"-c")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((CHROM_CONTAINER_SIZE += atoi(argv[i])) == 0) {
				fprintf(stderr, "ERROR: Chromosome Container Size must be an integer value unequal to 0!\n");
				exit(2);
			}
		}
		
		//seed threshold
		if(strcmp(argv[i],"-s")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((MAX_OCC_PER_SEED = atoi(argv[i])) == 0) {
				fprintf(stderr, "ERROR: Max occurrences per seed must be an integer value unequal to 0!\n");
				exit(2);
			}
		}

		//seed extend threshold
		if(strcmp(argv[i],"-x")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((SEED_EXTRA_EXTEND_THRESHOLD= atoi(argv[i])) == 0) {
				fprintf(stderr, "ERROR: High-frequent seed extension threshold must be an integer value unequal to 0!\n");
				exit(2);
			}
			if (SEED_EXTRA_EXTEND_THRESHOLD>DEFAULT_MAX_OCC_PER_SEED || SEED_EXTRA_EXTEND_THRESHOLD<0) {
				fprintf(stderr, "ERROR: High-frequent seed extension threshold must be positive and smaller than %d!\n",DEFAULT_MAX_OCC_PER_SEED);
				exit(2);
			}
		}
		
		//nr of unallowed chars in reads
		/*if(strcmp(argv[i],"-n")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((NUM_UNALLOWED_CHARS = atoi(argv[i])) == 0 && argv[i][0] != '0') {
				fprintf(stderr, "ERROR: Number of non-base symbols in read must be an integer value!\n");
				exit(2);
			}
		}*/
		
		//gaps most right
		if(strcmp(argv[i],"-d")==0){
			not_defined = 0;
			GAPS_MOST_RIGHT = 1;
		}
		
		//overhang alignment
		if(strcmp(argv[i],"-h")==0){
			not_defined = 0;
			OVERHANG_ALIGNMENT = 0;
		}
		
		//overhang alignment
		if(strcmp(argv[i],"-w")==0){
			not_defined = 0;
			STRINGENT_GAPLIMIT = 0;
		}
		
		//statistics
		//if(strcmp(argv[i],"-s")==0){
		//	not_defined = 0;
		//	STATISTICS = 1;
		//}
		
                //read offset (map only from j-th read on)
                if(strcmp(argv[i],"-j")==0){
                	not_defined = 0;
                	if (i+1 > argc - 1){ fprintf(stderr, "-j needs an positive integer value\n"); exit(2); }
                	i++;
                	if (((FROM_READ = atoi(argv[i])) == 0) || FROM_READ < 1) {
                		fprintf(stderr, "-j value must be a positive integer value!\n");
	                	exit(2);
			}
		}

		//print out gene, too (for every hit) used for WMD2
		if(strcmp(argv[i],"-target_info")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ fprintf(stderr, "-target_info needs an integer value (1: length only, 2: length+sequence)\n"); exit(2); }
			i++;
			if (((PRINT_SEQ = atoi(argv[i])) == 0) || PRINT_SEQ < 1 || PRINT_SEQ > 2) {
				fprintf(stderr, "-target_info value must be either 1 or 2!\n");
				exit(2);
			}
		}
		
		//flanking region of a hit
		if(strcmp(argv[i],"-flanking")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ fprintf(stderr, "-flanking needs an integer value, the length of one of the symmetric flanking regions of a hit\n"); exit(2); }
			i++;
			if (((FLANKING = atoi(argv[i])) == 0) || FLANKING < 0 || FLANKING > 100) {
				fprintf(stderr, "-flanking value must be a positive integer and must not exceed 100!\n");
				exit(2);
			}
		}

		// debug mode
		if(strcmp(argv[i],"-debug")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ DEBUGREAD = 0; DEBUG = 1; }
			else {
				i++;
				if ((DEBUGREAD = atoi(argv[i])) == 0) DEBUGREAD = 0;
			}
		}

#ifdef THREADS
		//number of threads 
		if(strcmp(argv[i],"-t")==0){
			not_defined = 0;
			if (i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((NUM_THREADS = atoi(argv[i])) == 0) {
				if (argv[i][0] != '0') {
					fprintf(stderr, "ERROR: Number of threads must be an integer value!\n");
					exit(2);
				}
			}
		}
#endif
		
		if (not_defined == 1) {
			usage(); exit(1);
		}
	}

	if (has_index == 0 || has_query == 0 || has_genome == 0) {
		usage(); exit(1);
	}
	
	if (REPEATMAP != 0 && (ALL_HIT_STRATEGY || SECOND_BEST_HIT_STRATEGY)) {
		fprintf(stderr, "ERROR: All hit/second best hit strategy and a limit on the output alignments exclude each other (options -a or -b and -n)!\n");
		exit(2);
	}

	if (SPLICED_HIT_MAX_DIST < 0) {
		fprintf(stderr, "ERROR: Spliced must have a positive distance\n");
		exit(1);
	}

	if (ALL_HIT_STRATEGY && SECOND_BEST_HIT_STRATEGY) {
		fprintf(stderr, "ERROR: All hit strategy and second best hit strategy exclude each other (options -a and -b)!\n");
		exit(1);
	}

	if (NUM_MISMATCHES > MAX_EDIT_OPS) {
		fprintf(stderr, "ERROR: Number of allowed mismatches exceeds maximal number of edit operations (=%d)! Please restart with a lower value!\n", MAX_EDIT_OPS);
		exit(2);
	}

	if (NUM_GAPS > MAX_EDIT_OPS) {
		fprintf(stderr, "ERROR: Number of allowed gaps exceeds maximal number of edit operations (=%d)! Please restart with a lower value!\n", MAX_EDIT_OPS);
		exit(2);
	}

	return 0;
}

int init_constants() 
{
	if (INDEX_DEPTH == 5) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 256;		//binary number: 0000 0000 0000 0001 0000 0000
		BINARY_CODE[2] = 512;		//binary number: 0000 0000 0000 0010 0000 0000
		BINARY_CODE[3] = 768;		//binary number: 0000 0000 0000 0011 0000 0000
	}
	if (INDEX_DEPTH == 6) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 1024;		//binary number: 0000 0000 0000 0100 0000 0000
		BINARY_CODE[2] = 2048;		//binary number: 0000 0000 0000 1000 0000 0000
		BINARY_CODE[3] = 3072;		//binary number: 0000 0000 0000 1100 0000 0000
	}
	if (INDEX_DEPTH == 7) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 4096;		//binary number: 0000 0000 0001 0000 0000 0000
		BINARY_CODE[2] = 8192;		//binary number: 0000 0000 0010 0000 0000 0000
		BINARY_CODE[3] = 12288;		//binary number: 0000 0000 0011 0000 0000 0000
	}
        if (INDEX_DEPTH == 8) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 16384;         //binary number: 0000 0000 0100 0000 0000 0000
                BINARY_CODE[2] = 32768;         //binary number: 0000 0000 1000 0000 0000 0000
                BINARY_CODE[3] = 49152;         //binary number: 0000 0000 1100 0000 0000 0000
        }
        if (INDEX_DEPTH == 9) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 65536;         //binary number: 0000 0001 0000 0000 0000 0000
                BINARY_CODE[2] = 131072;        //binary number: 0000 0010 0000 0000 0000 0000
                BINARY_CODE[3] = 196608;        //binary number: 0000 0011 0000 0000 0000 0000
        }
        if (INDEX_DEPTH == 10) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 262144;        //binary number: 0000 0100 0000 0000 0000 0000
                BINARY_CODE[2] = 524288;        //binary number: 0000 1000 0000 0000 0000 0000
                BINARY_CODE[3] = 786432;        //binary number: 0000 1100 0000 0000 0000 0000
        }
        if (INDEX_DEPTH == 11) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 1048576;       //binary number: 0001 0000 0000 0000 0000 0000
                BINARY_CODE[2] = 2097152;       //binary number: 0010 0000 0000 0000 0000 0000
                BINARY_CODE[3] = 3145728;       //binary number: 0011 0000 0000 0000 0000 0000
        }
        if (INDEX_DEPTH == 12) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 4194304;       //binary number: 0100 0000 0000 0000 0000 0000
                BINARY_CODE[2] = 8388608;       //binary number: 1000 0000 0000 0000 0000 0000
                BINARY_CODE[3] = 12582912;      //binary number: 1100 0000 0000 0000 0000 0000
        }
        if (INDEX_DEPTH == 13) {
                BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000 0000
                BINARY_CODE[1] = 16777216;      //binary number: 0001 0000 0000 0000 0000 0000 0000
                BINARY_CODE[2] = 33554432;      //binary number: 0010 0000 0000 0000 0000 0000 0000
                BINARY_CODE[3] = 50331648;      //binary number: 0011 0000 0000 0000 0000 0000 0000
        }

	return(0);
}

int init_output_file() 
{
	if (strlen(OUT_FILE_NAME) > 0) {
		if ((OUT_FP = fopen(OUT_FILE_NAME, "w")) == NULL) {
			perror("ERROR : Couldn't open output file");
			fprintf(stderr, "ERROR : %s\n", OUT_FILE_NAME);
			exit(1);
		}
	} else {
		OUT_FP = stdout;
	}

	return(0);
}

int init_spliced_output_file()
{
	if (SPLICED_HITS && strlen(SPLICED_OUT_FILE_NAME) > 0) {
                if ((SP_OUT_FP = fopen(SPLICED_OUT_FILE_NAME, "w")) == NULL) {
                        perror("ERROR : Couldn't open output file for spliced hits");
                        fprintf(stderr, "ERROR : %s\n", SPLICED_OUT_FILE_NAME);
                        exit(1);
                }
        } else {
                SP_OUT_FP = stdout;
        }

        return(0);
}

int init_leftover_file() 
{
	if ((LEFTOVER_FP = fopen(LEFTOVER_FILE_NAME, "w")) == NULL) {
		perror("ERROR : Couldn't open leftover file");
		fprintf(stderr, "ERROR : %s\n", LEFTOVER_FILE_NAME);
		exit(1);
	}

	return(0);
}

int init_query_file() 
{
	if ((QUERY_FP = fopen(QUERY_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open input file");
		fprintf(stderr, "ERROR : %s\n", QUERY_FILE_NAME);
		exit(1);
	}

	return(0);
}

int init_chr_index_file() 
{
	if ((CHR_INDEX_FP = fopen(CHR_INDEX_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open input file");
		fprintf(stderr, "ERROR : %s\n", CHR_INDEX_FILE_NAME);
		exit(1);
	}

	return(0);
}

int init_meta_index_file() 
{

#ifndef METHYLOME
	if ((META_INDEX_FP = fopen(META_INDEX_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open input file");
		fprintf(stderr, "ERROR : %s\n", META_INDEX_FILE_NAME);
		exit(1);
	}
#else
	if ((META_INDEX_CT_FP = fopen(META_INDEX_CT_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open input file");
                fprintf(stderr, "ERROR : %s\n", META_INDEX_CT_FILE_NAME);
                exit(1);
        }
	if ((META_INDEX_GA_FP = fopen(META_INDEX_GA_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open input file");
                fprintf(stderr, "ERROR : %s\n", META_INDEX_GA_FILE_NAME);
                exit(1);
        }
#endif


	return(0);
}

int init_genome_file() 
{
	if ((GENOME_FP = fopen(GENOME_FILE_NAME, "r")) == NULL) {
		perror("ERROR : Couldn't open genome file");
		fprintf(stderr, "ERROR : %s\n", GENOME_FILE_NAME);
		exit(1);
	}

	return 0;
}


int init_hit_lists() 
{
	alloc_hit_lists_operator();	
	alloc_hits_by_score();	// A T T E N T I O N ! ! !   needs correct NUM_EDIT_OPS which can be changed in init_alignment_structures() !!!

	return(0);
}

int init_operators()
{
	MAPPING_ENTRY_OPERATOR = alloc_mapping_entry_container();
	MAPPING_ENTRY_OPERATOR_FIRST = MAPPING_ENTRY_OPERATOR;

	HIT_OPERATOR = alloc_hit_container();
	HIT_OPERATOR_FIRST = HIT_OPERATOR;

	alloc_chromosome_entry_container();

	return(0);
}

/**
int init_statistic_vars()
{
        int i;
        PERFECT_READS = 0;
        PERFECT_HITS = 0;
        PERFECT_HITS_REV = 0;
        for (i=0; i!=37; ++i) HITS_LEN[i] = 0;
        HITS_MM = (unsigned int *) malloc ((MAX_EDIT_OPS + 1) * sizeof(unsigned int));
        for (i=0; i!=MAX_EDIT_OPS+1; ++i) HITS_MM[i] = 0;
        NUM_ALIGNMENTS = 0;
        NUM_WHOLE_ALIGNMENTS = 0;
        ENDSTART_MAPPED[0] = 0;
        ENDSTART_MAPPED[1] = 0;
        NOT_ALIGNED[0] = 0;
        NOT_ALIGNED[1] = 0;
        GAPS_ENCOUNTERED[0] = 0;
        GAPS_ENCOUNTERED[1] = 0;
        GAPS_ENCOUNTERED[2] = 0;
        TOO_MANY_MMS[0] = 0;
        TOO_MANY_MMS[1] = 0;
        BREAK_TB_IN_GLOBAL_ALIGNMENT = 0;
        BREAK_GLOBAL_ALIGNMENT[0] = 0;
        BREAK_GLOBAL_ALIGNMENT[1] = 0;
        CELLS_GLOBAL = 0;
        CELLS_OVERHANG = 0;

        W = 0;

        return(0);
}
*/

